@@ -1,4 +1,4 @@
-# Copyright (C) 2012-2019  Ruby-GNOME2 Project Team
+# Copyright (C) 2012-2019  Ruby-GNOME Project Team
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -80,6 +80,10 @@ module GObjectIntrospection
       define_singleton_method(@base_module, name, info)
     end
 
+    def load_function_info_singleton_method(info, klass, method_name)
+      define_singleton_method(klass, method_name, info)
+    end
+
     def define_module_function(target_module, name, function_info)
       function_info.unlock_gvl = should_unlock_gvl?(function_info, target_module)
       full_method_name = "#{target_module}\#.#{name}"
@@ -125,7 +129,7 @@ module GObjectIntrospection
         end
         unless method_infos.empty?
           base_class = @base_module.const_get(base_class_name)
-          load_method_infos(method_infos, base_class.singleton_class)
+          load_methods_method(method_infos, base_class.singleton_class)
         end
       else
         return if info.gtype_struct?
@@ -135,6 +139,7 @@ module GObjectIntrospection
     end
 
     def define_boxed(info)
+      return if info.gtype == GLib::Type::NONE
       klass = self.class.define_class(info.gtype, info.name, @base_module)
       _ = klass # TODO: Remove me. It is just for suppressing a warning.
       # TODO
@@ -151,6 +156,8 @@ module GObjectIntrospection
     end
 
     def define_enum(info)
+      # TODO: Can we do the same things for flags on NONE GType?
+      return if info.gtype == GLib::Type::NONE
       self.class.define_class(info.gtype,
                               rubyish_class_name(info),
                               @base_module)
@@ -205,6 +212,7 @@ module GObjectIntrospection
     end
 
     def load_object_info(info)
+      return if info.gtype == GLib::Type::NONE
       klass = self.class.define_class(info.gtype,
                                       rubyish_class_name(info),
                                       @base_module)
@@ -273,11 +281,11 @@ module GObjectIntrospection
         next if method_infos.empty?
         case method_infos.first
         when ConstructorInfo
-          load_constructor_infos(method_infos, klass)
+          load_methods_constructor(method_infos, klass)
         when MethodInfo
-          load_method_infos(method_infos, klass)
+          load_methods_method(method_infos, klass)
         when FunctionInfo
-          load_function_infos(method_infos, klass)
+          load_methods_function(method_infos, klass)
         else
           raise "TODO: #{method_info_class}"
         end
@@ -288,16 +296,18 @@ module GObjectIntrospection
       end
     end
 
-    def load_constructor_infos(infos, klass)
+    def load_methods_constructor(infos, klass)
       return if infos.empty?
 
       call_initialize_post = lambda do |object|
         initialize_post(object)
       end
+      invokers = []
       infos.each do |info|
         name = "initialize_#{info.name}"
         info.unlock_gvl = should_unlock_gvl?(info, klass)
         invoker = Invoker.new(info, name, "#{klass}\##{name}")
+        invokers << invoker
         klass.__send__(:define_method, name) do |*arguments, &block|
           invoker.invoke(self, arguments, block)
           call_initialize_post.call(self)
@@ -305,12 +315,26 @@ module GObjectIntrospection
         klass.__send__(:private, name)
       end
 
-      find_info = lambda do |arguments|
-        find_suitable_callable_info(infos, arguments)
+      initialize = lambda do |receiver, arguments, block|
+        invokers.each do |invoker|
+          catch do |tag|
+            invoker.invoke(receiver, arguments.dup, block, tag)
+            call_initialize_post.call(receiver)
+            return
+          end
+        end
+        message = "wrong arguments: "
+        message << "#{klass.name}#initialize("
+        message << arguments.collect(&:inspect).join(", ")
+        message << "): "
+        message << "available signatures"
+        invokers.each do |invoker|
+          message << ": #{invoker.signature}"
+        end
+        raise ArgumentError, message
       end
       klass.__send__(:define_method, "initialize") do |*arguments, &block|
-        info = find_info.call(arguments, &block)
-        __send__("initialize_#{info.name}", *arguments, &block)
+        initialize.call(self, arguments, block)
       end
     end
 
@@ -502,7 +526,7 @@ module GObjectIntrospection
       false
     end
 
-    def load_method_infos(infos, klass)
+    def load_methods_method(infos, klass)
       infos.each do |info|
         method_name = rubyish_method_name(info)
         load_method_info(info, klass, method_name)
@@ -546,16 +570,17 @@ module GObjectIntrospection
       end
     end
 
-    def load_function_infos(infos, klass)
+    def load_methods_function(infos, klass)
       infos.each do |info|
         name = rubyish_method_name(info)
         next if name == "new"
         next if name == "alloc"
-        define_singleton_method(klass, name, info)
+        load_function_info_singleton_method(info, klass, name)
       end
     end
 
     def load_interface_info(info)
+      return if info.gtype == GLib::Type::NONE
       interface_module =
         self.class.define_interface(info.gtype,
                                     rubyish_class_name(info),
@@ -572,6 +597,7 @@ module GObjectIntrospection
     end
 
     def load_union_info(info)
+      return if info.gtype == GLib::Type::NONE
       klass = self.class.define_class(info.gtype, info.name, @base_module)
       load_fields(info, klass)
       load_methods(info, klass)
@@ -585,7 +611,7 @@ module GObjectIntrospection
         @prepared = false
       end
 
-      def invoke(receiver, arguments, block)
+      def invoke(receiver, arguments, block, abort_tag=nil)
         ensure_prepared
 
         if receiver and @function_info_p
@@ -593,13 +619,16 @@ module GObjectIntrospection
         end
 
         arguments, block = build(receiver, arguments, block)
-        unless valid?(arguments)
-          if @on_invalid == :fallback
-            return @value_on_invalid
+        if wrong_number_of_arguments?(arguments)
+          if abort_tag
+            throw(abort_tag)
           else
-            raise ArgumentError, invalid_error_message(invoke_arguments)
+            raise ArgumentError, invalid_error_message(arguments)
           end
         end
+        unless normalize_arguments!(arguments, abort_tag)
+          return @value_on_invalid
+        end
 
         if block.nil? and @require_callback_p
           receiver.to_enum(@method_name, *arguments)
@@ -617,6 +646,12 @@ module GObjectIntrospection
         end
       end
 
+      def signature
+        ensure_prepared
+        argument_signatures = @in_args.collect(&:signature)
+        "(" + argument_signatures.join(", ") + ")"
+      end
+
       private
       def ensure_prepared
         return if @prepared
@@ -678,17 +713,31 @@ module GObjectIntrospection
         return arguments, block
       end
 
-      def valid?(arguments)
-        return false unless @valid_n_args_range.cover?(arguments.size)
-        if @on_invalid == :fallback
-          arguments.zip(@in_arg_types) do |argument, type|
-            return false unless type.match?(argument)
+      def wrong_number_of_arguments?(arguments)
+        not @valid_n_args_range.cover?(arguments.size)
+      end
+
+      def normalize_arguments!(arguments, abort_tag)
+        arguments.size.times do |i|
+          argument = arguments[i]
+          next if argument.nil?
+          type = @in_arg_types[i]
+          converted_argument = type.try_convert(argument)
+          if converted_argument.nil?
+            if abort_tag
+              throw(abort_tag)
+            elsif @on_invalid == :fallback
+              return false
+            else
+              next
+            end
           end
+          arguments[i] = converted_argument
         end
         true
       end
 
-      def invalid_error_message(argumetns)
+      def invalid_error_message(arguments)
         detail = "#{arguments.size} for "
         if @n_in_args == @n_required_in_args
           detail << "#{@n_in_args}"
